@@ -4,6 +4,7 @@ import copy
 import difflib
 import lxml
 import re
+import sys
 import unittest
 import xml2rfc
 import xml2rfc.utils
@@ -11,7 +12,10 @@ import xml2rfc.utils
 from xml2rfc.boilerplate_rfc_7841 import boilerplate_rfc_status_of_memo
 from xml2rfc.walkpdf import xmldoc
 from xml2rfc.writers.base import default_options, BaseV3Writer, RfcWriterError
+from xml2rfc.writers import DatatrackerToBibConverter
 from xml2rfc.writers.text import MAX_WIDTH
+from xml2rfc.utils import strip_link_attachments
+from xml2rfc.util.file import can_access, FileAccessError
 
 try:
     from xml2rfc import debug
@@ -468,6 +472,7 @@ class WriterRfcTest(WriterRootTest):
         return self.status_test()
 
 
+@unittest.skipIf(sys.platform.startswith("win"), "Test skipped on Windows OS")
 class PdfWriterTests(unittest.TestCase):
     elements_root = None
     elements_pdfxml = None
@@ -697,6 +702,78 @@ class TextWriterTest(unittest.TestCase):
         self.assertEqual(len(lines), 2)
         self.assertIn(url, lines[1].text)
 
+    def test_render_artwork_svg_only(self):
+        artwork_str = '''
+<artwork type="svg">
+    <svg xmlns="http://www.w3.org/2000/svg">
+      <text x="10" y="15">foobar</text>
+    </svg>
+</artwork>'''
+        artwork = lxml.etree.fromstring(artwork_str)
+
+        # for RFC
+        self.writer.root = lxml.etree.fromstring(f'<rfc number="1234"></rfc>')
+        self.writer.options.rfc = True
+        self.writer.rfcnumber = 1234
+        lines = self.writer.render_artwork(artwork, width=60)
+        text = ''.join(line.text for line in lines)
+        self.assertIn("Artwork only available as SVG", text)
+        self.assertIn(default_options.rfc_html_archive_url, text)
+        self.assertIn('rfc1234.html', text)
+
+        # for Draft
+        self.writer.options.rfc = None
+        lines = self.writer.render_artwork(artwork, width=60)
+        text = ' '.join(line.text for line in lines)
+        self.assertIn("Artwork only available as SVG", text)
+        self.assertIn(default_options.id_html_archive_url, text)
+        self.assertIn(".html", text)
+
+    def test_render_artwork_binary_art_only(self):
+        artwork_str = '<artwork type="binary-art" src="foot.bin" />'
+        artwork = lxml.etree.fromstring(artwork_str)
+
+        # for RFC
+        self.writer.root = lxml.etree.fromstring(f'<rfc number="1234"></rfc>')
+        self.writer.options.rfc = True
+        self.writer.rfcnumber = 1234
+        lines = self.writer.render_artwork(artwork, width=60)
+        text = ''.join(line.text for line in lines)
+        self.assertIn("Artwork only available as BINARY-ART", text)
+        self.assertIn(default_options.rfc_html_archive_url, text)
+        self.assertIn('rfc1234.html', text)
+
+        # for Draft
+        self.writer.options.rfc = None
+        lines = self.writer.render_artwork(artwork, width=60)
+        text = ' '.join(line.text for line in lines)
+        self.assertIn("Artwork only available as BINARY-ART", text)
+        self.assertIn(default_options.id_html_archive_url, text)
+        self.assertIn(".html", text)
+
+    def test_render_artwork_non_ascii_unknown(self):
+        artwork_str = '<artwork src="foot.bin" />'
+        artwork = lxml.etree.fromstring(artwork_str)
+
+        # for RFC
+        self.writer.root = lxml.etree.fromstring(f'<rfc number="1234"></rfc>')
+        self.writer.options.rfc = True
+        self.writer.rfcnumber = 1234
+        lines = self.writer.render_artwork(artwork, width=60)
+        text = ''.join(line.text for line in lines)
+        self.assertIn("Artwork only available as (unknown type)", text)
+        self.assertIn(default_options.rfc_html_archive_url, text)
+        self.assertIn('rfc1234.html', text)
+
+        # for Draft
+        self.writer.options.rfc = None
+        lines = self.writer.render_artwork(artwork, width=60)
+        text = ' '.join(line.text for line in lines)
+        self.assertIn("Artwork only available as (unknown type)", text)
+        self.assertIn(default_options.id_html_archive_url, text)
+        self.assertIn(".html", text)
+
+
 class BaseV3WriterTest(unittest.TestCase):
     '''BaseV3Writer tests'''
 
@@ -772,6 +849,258 @@ class BaseV3WriterTest(unittest.TestCase):
         with self.assertRaises(RfcWriterError):
             self.writer.validate_draft_name()
 
+    def test_get_relavent_pis_root(self):
+        xml = lxml.etree.fromstring(
+            """
+<rfc
+    ipr="trust200902"
+    submissionType="editorial"
+    category="info">
+    <?v3xml2rfc table_borders="light" ?>
+    <front>
+        <section>
+          <table>
+            <thead>
+              <tr>
+                <th>Column 1</th>
+                <th>Column 2</th>
+              </tr>
+            </thead>
+            <tbody>
+              <tr>
+                <td>Value 1</td>
+                <td>Value 2</td>
+              </tr>
+            </tbody>
+          </table>
+        </section>
+    </front>
+</rfc>"""
+        )
+        self.writer.root = xml
+        table = self.writer.root.xpath("//table")[0]
+        self.assertEqual(self.writer.get_relevant_pi(table, "table_borders"), "light")
+
+    def test_get_relavent_pis_before(self):
+        xml = lxml.etree.fromstring(
+            """
+<rfc
+    ipr="trust200902"
+    submissionType="editorial"
+    category="info">
+    <?v3xml2rfc table_borders="light" ?>
+    <front>
+        <section>
+        <?v3xml2rfc table_borders="full" ?>
+          <table>
+            <thead>
+              <tr>
+                <th>Column 1</th>
+                <th>Column 2</th>
+              </tr>
+            </thead>
+            <tbody>
+              <tr>
+                <td>Value 1</td>
+                <td>Value 2</td>
+              </tr>
+            </tbody>
+          </table>
+        </section>
+    </front>
+</rfc>"""
+        )
+        self.writer.root = xml
+        table = self.writer.root.xpath("//table")[0]
+        self.assertEqual(self.writer.get_relevant_pi(table, "table_borders"), "full")
+
+    def test_get_relavent_pis_inside(self):
+        xml = lxml.etree.fromstring(
+            """
+<rfc
+    ipr="trust200902"
+    submissionType="editorial"
+    category="info">
+    <?v3xml2rfc table_borders="light" ?>
+    <front>
+        <section>
+        <?v3xml2rfc table_borders="full" ?>
+          <table>
+            <?v3xml2rfc table_borders="min" ?>
+            <thead>
+              <tr>
+                <th>Column 1</th>
+                <th>Column 2</th>
+              </tr>
+            </thead>
+            <tbody>
+              <tr>
+                <td>Value 1</td>
+                <td>Value 2</td>
+              </tr>
+            </tbody>
+          </table>
+        </section>
+    </front>
+</rfc>"""
+        )
+        self.writer.root = xml
+        table = self.writer.root.xpath("//table")[0]
+        self.assertEqual(self.writer.get_relevant_pi(table, "table_borders"), "min")
+
+    def test_get_relavent_pis_siblings(self):
+        xml = lxml.etree.fromstring(
+            """
+<rfc
+    ipr="trust200902"
+    submissionType="editorial"
+    category="info">
+    <?v3xml2rfc table_borders="light" ?>
+    <front>
+        <section>
+          <?v3xml2rfc table_borders="full" ?>
+          <table>
+            <thead>
+              <tr>
+                <th>Column 1</th>
+              </tr>
+            </thead>
+            <tbody>
+              <tr>
+                <td>Value 1</td>
+              </tr>
+            </tbody>
+          </table>
+          <?v3xml2rfc table_borders="min" ?>
+          <table>
+            <thead>
+              <tr>
+                <th>Column 2</th>
+              </tr>
+            </thead>
+            <tbody>
+              <tr>
+                <td>Value 2</td>
+              </tr>
+            </tbody>
+          </table>
+        </section>
+    </front>
+</rfc>"""
+        )
+        self.writer.root = xml
+        table_1 = self.writer.root.xpath("//table")[0]
+        self.assertEqual(self.writer.get_relevant_pi(table_1, "table_borders"), "full")
+        table_2 = self.writer.root.xpath("//table")[1]
+        self.assertEqual(self.writer.get_relevant_pi(table_2, "table_borders"), "min")
+
+
+class DatatrackerToBibConverterTest(unittest.TestCase):
+    """DatatrackerToBibConverter tests"""
+
+    def setUp(self):
+        xml2rfc.log.quiet = True
+        path = "tests/input/elements.xml"
+        self.parser = xml2rfc.XmlRfcParser(path,
+                                           quiet=True,
+                                           options=default_options,
+                                           **options_for_xmlrfcparser)
+        self.xmlrfc = self.parser.parse()
+        self.writer = xml2rfc.DatatrackerToBibConverter(self.xmlrfc, quiet=True)
+
+    def test_convert(self):
+        reference_a = "reference.I-D.ietf-sipcore-multiple-reasons.xml"
+        reference_b = "reference.I-D.draft-ietf-sipcore-multiple-reasons-01.xml"
+        rfc = lxml.etree.fromstring(f"""
+<rfc xmlns:xi="http://www.w3.org/2001/XInclude">
+    <xi:include href="https://datatracker.ietf.org/doc/bibxml3/{reference_a}"/>
+    <xi:include href="https://datatracker.ietf.org/doc/bibxml3/{reference_b}"/>
+</rfc>""")
+        self.writer.root = rfc
+        self.writer.convert_xincludes()
+
+        ns = {"xi": b"http://www.w3.org/2001/XInclude"}
+        xincludes = self.writer.root.xpath("//xi:include", namespaces=ns)
+
+        # reference without revision
+        self.assertEqual(xincludes[0].get("href"), f"https://bib.ietf.org/public/rfc/bibxml-ids/{reference_a}")
+
+        # reference with revision
+        self.assertEqual(xincludes[1].get("href"), f"https://bib.ietf.org/public/rfc/bibxml-ids/{reference_b}")
+
+
+class FileAccessTest(unittest.TestCase):
+    def setUp(self):
+        self.options = copy.deepcopy(default_options)
+
+    def test_allow_access_false(self):
+        self.options.allow_local_file_access = False
+        with self.assertRaises(FileAccessError) as e:
+            can_access(self.options, "/foo/bar", "/bar")
+            self.assertIn("Can not access local file" in e.exception)
+
+    def test_has_meta_chars(self):
+        self.options.allow_local_file_access = True
+        for char in "[><*[`$|;&(#]":
+            with self.assertRaises(FileAccessError) as e:
+                can_access(self.options, "/foo/bar", f"/bar{char}")
+                self.assertIn("Found disallowed shell meta-characters" in e.exception)
+
+    def test_not_within_src_dir(self):
+        self.options.allow_local_file_access = True
+        with self.assertRaises(FileAccessError) as e:
+            can_access(self.options, "/foo/bar", "/etc/passwd")
+            self.assertIn("Expected a file located beside or below the .xml source" in e.exception)
+
+    def test_non_existent_file(self):
+        self.options.allow_local_file_access = True
+        with self.assertRaises(FileAccessError) as e:
+            can_access(self.options, "test.py", "foobar.foobar")
+            self.assertIn("no such file exists" in e.exception)
+
+    def test_allowed_access(self):
+        self.options.allow_local_file_access = True
+        self.assertTrue(can_access(self.options, "test.py", "test.py"))
+
+    def test_allowed_access_template(self):
+        self.options.allow_local_file_access = False
+        self.assertTrue(can_access(self.options,
+                                   source="test.py",
+                                   path="rfc2629-xhtml.ent",
+                                   access_templates=True))
+
+
+class StripLinkAttachmentsTest(unittest.TestCase):
+    def setUp(self):
+        self.options = copy.deepcopy(default_options)
+
+    def test_strips_attachement(self):
+        self.options.allow_local_file_access = False
+        xml_tree = lxml.etree.fromstring("""
+<rfc>
+    <link rel="item" href="urn:issn:0000-0000" />
+    <link rel="attachment" href="/etc/passwd" />
+    <link rel="describedBy" href="doi:00.00000/rfc0000" />
+    <link rel="prev" href="https://www.ietf.org/archive/id/draft-ietf-empire-death-star-00.txt" />
+    <link rel="attachment" href="/home/vader/ds-1/schematics" />
+    <link rel="alternate" href="https://www.ietf.org/ds-1" />
+</rfc>""")
+
+        strip_link_attachments(xml_tree)
+
+        # keeps non rel="attachment" elements
+        for rel in ["item", "describedBy", "prev", "alternate"]:
+            links = xml_tree.xpath(f'//link[@rel="{rel}"]')
+            self.assertEqual(len(links), 1)
+
+        # strip rel="attachment"
+        attachments = xml_tree.xpath('//link[@rel="attachment"]')
+        self.assertEqual(len(attachments), 0)
+        xml_str = lxml.etree.tostring(xml_tree).decode()
+        self.assertNotIn("attachment", xml_str)
+        self.assertNotIn("/etc/passwd", xml_str)
+        self.assertNotIn("/home/vader/ds-1/schematics", xml_str)
+
 
 if __name__ == '__main__':
     unittest.main()
